from c3.signal.pulse import Envelope, EnvelopeNetZero
from c3.c3objs import Quantity as Qty
from c3.libraries import envelopes
import numpy as np
import pytest

flux_env = Envelope(
    name="flux",
    desc="Flux bias for tunable coupler",
    shape=envelopes.rect,
    params={
        "amp": Qty(value=0.1, unit="V"),
        "t_final": Qty(value=10e-9, unit="s"),
        "freq_offset": Qty(value=0, min_val=0, max_val=1, unit="Hz 2pi"),
        "xy_angle": Qty(value=0, min_val=0, max_val=np.pi, unit="rad"),
    },
)

flux_env_netzero = EnvelopeNetZero(
    name="flux",
    desc="Flux bias for tunable coupler",
    shape=envelopes.rect,
    params={
        "amp": Qty(value=0.1, unit="V"),
        "t_final": Qty(value=5e-9, unit="s"),
        "freq_offset": Qty(value=0, min_val=0, max_val=1, unit="Hz 2pi"),
        "xy_angle": Qty(value=0, min_val=0, max_val=np.pi, unit="rad"),
    },
)


@pytest.mark.unit
def test_envelope() -> None:
    ts = np.arange(0, 12e-9, 1e-9)
    shape = flux_env.get_shape_values(ts)
    assert np.all(
        shape.numpy()
        == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0])
    )


@pytest.mark.unit
def test_envelope_netzero() -> None:
    ts = np.arange(0, 12e-9, 1e-9)
    shape = flux_env_netzero.get_shape_values(ts)
    assert np.all(
        shape.numpy()
        == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 0.0, -1.0, -1.0, -1.0, -1.0, -1.0, -0.0])
    )



from c3.c3objs import C3obj, hjson_encode
from c3.c3objs import Quantity as Qty
from c3.libraries.envelopes import envelopes
from c3.utils.tf_utils import tf_complexify
import tensorflow as tf
import numpy as np
import types
import hjson
from typing import Callable, Union, Dict

components = dict()


def comp_reg_deco(func):
    """
    Decorator for making registry of functions
    """
    components[str(func.__name__)] = func
    return func


@comp_reg_deco
class Envelope(C3obj):
    """
    Represents the envelopes shaping a pulse.

    Parameters
    ----------
    shape: Callable
        function evaluating the shape in time

    """

    def __init__(
        self,
        name: str,
        desc: str = " ",
        comment: str = " ",
        params: Dict[str, Qty] = {},
        shape: Union[Callable, str] = None,
        use_t_before=False,
    ):
        if isinstance(shape, str):
            self.shape = envelopes[shape]
        else:
            self.shape = shape
        default_params = {
            "amp": Qty(value=1.0, min_val=-1.0, max_val=+1.5, unit="V"),
            "delta": Qty(value=0.0, min_val=-5.0, max_val=+5.0, unit="V"),
            "freq_offset": Qty(value=0.0, min_val=-1.0, max_val=+1.0, unit="Hz 2pi"),
            "xy_angle": Qty(value=0.0, min_val=-1.0, max_val=+1.0, unit="rad"),
            "sigma": Qty(value=5e-9, min_val=-2.0, max_val=+2.0, unit="s"),
            "t_final": Qty(value=1.0, min_val=-1.0, max_val=+1.0, unit="s"),
        }
        default_params.update(params)
        self.set_use_t_before(use_t_before)
        super().__init__(
            name=name,
            desc=desc,
            comment=comment,
            params=default_params,
        )

    def write_config(self, filepath: str) -> None:
        """
        Write dictionary to a HJSON file.
        """
        with open(filepath, "w") as cfg_file:
            hjson.dump(self.asdict(), cfg_file, default=hjson_encode)

    def asdict(self) -> dict:
        params = {}
        for key, item in self.params.items():
            params[key] = item.asdict()
        return {
            "name": self.name,
            "c3type": self.__class__.__name__,
            "shape": self.shape.__name__,
            "params": params,
        }

    def __str__(self) -> str:
        return hjson.dumps(self.asdict(), default=hjson_encode)

    def __repr__(self) -> str:
        repr_str = self.name + ":: "
        for key, item in self.params.items():
            repr_str += str(key) + " : " + str(item) + ", "
        repr_str += "shape: " + self.shape.__name__ + ", "
        return repr_str

    def set_use_t_before(self, use_t_before):
        if use_t_before:
            self.get_shape_values = self._get_shape_values_before
        else:
            self.get_shape_values = self._get_shape_values_just

    def compute_mask(self, ts, t_end) -> tf.Tensor:
        """Compute a mask to cut out a signal after t_final.

        Parameters
        ----------
        ts : tf.Tensor
            Vector of time steps.

        Returns
        -------
        tf.Tensor
            [description]
        """
        t_final = tf.minimum(self.params["t_final"].get_value(), t_end)
        dt = ts[1] - ts[0]
        return tf_complexify(
            tf.sigmoid((ts / dt + 0.001) * 1e6)
            * tf.sigmoid((0.999 * t_final - ts) / dt * 1e6)
        )

    def _get_shape_values_before(self, ts, t_final=1):
        """Return the value of the shape function at the specified times. With the offset, we make sure the
        signal starts with amplitude zero by subtracting the shape value at time -dt.

        Parameters
        ----------
        ts : tf.Tensor
            Vector of time samples.
        """
        t_before = 2 * ts[0] - ts[1]  # t[0] - (t[1] - t[0])
        offset = self.shape(t_before, self.params)
        mask = self.compute_mask(ts, t_final)
        return mask * (self.shape(ts, self.params) - offset)

    def _get_shape_values_just(self, ts, t_final=1):
        """Return the value of the shape function at the specified times.

        Parameters
        ----------
        ts : tf.Tensor
            Vector of time samples.
        """
        mask = self.compute_mask(ts, t_final)
        return mask * self.shape(ts, self.params)


@comp_reg_deco
class EnvelopeDrag(Envelope):
    def __init__(
        self,
        name: str,
        desc: str = " ",
        comment: str = " ",
        params: dict = {},
        shape: types.FunctionType = None,
        use_t_before=False,
    ):
        super().__init__(
            name=name,
            desc=desc,
            comment=comment,
            params=params,
            shape=shape,
            use_t_before=use_t_before,
        )
        self.set_use_t_before(use_t_before)

    def set_use_t_before(self, use_t_before):
        if use_t_before:
            self.base_env = super()._get_shape_values_before
        else:
            self.base_env = super()._get_shape_values_just

    def get_shape_values(self, ts, t_final=1):
        dt = ts[1] - ts[0]
        with tf.GradientTape() as t:
            t.watch(ts)
            env = tf.math.real(self.base_env(ts, t_final))
        denv = (
            t.gradient(env, ts, unconnected_gradients=tf.UnconnectedGradients.ZERO) * dt
        )  # Derivative W.R.T. to bins
        delta = self.params["delta"].get_value()
        return tf.complex(env, -denv * delta)


@comp_reg_deco
class EnvelopeNetZero(Envelope):
    """
    Represents the envelopes shaping a pulse.

    Parameters
    ----------
    shape: function
        function evaluating the shape in time
    params: dict
        Parameters of the envelope
        Note: t_final
    """

    def __init__(
        self,
        name: str,
        desc: str = " ",
        comment: str = " ",
        params: dict = {},
        shape: types.FunctionType = None,
        use_t_before=False,
    ):
        super().__init__(
            name=name,
            desc=desc,
            comment=comment,
            params=params,
            shape=shape,
            use_t_before=use_t_before,
        )
        self.set_use_t_before(use_t_before)

    def set_use_t_before(self, use_t_before):
        if use_t_before:
            self.base_env = super()._get_shape_values_before
        else:
            self.base_env = super()._get_shape_values_just

    def get_shape_values(self, ts):
        """Return the value of the shape function at the specified times.

        Parameters
        ----------
        ts : tf.Tensor
            Vector of time samples.
        t_before : tf.float64
            Offset the beginning of the shape by this time.
        """
        N_red = len(ts) // 2
        ts_red = tf.split(ts, [N_red, len(ts) - N_red], 0)[0]
        shape_values = self.base_env(ts=ts_red)
        netzero_shape_values = tf.concat(
            [shape_values, -shape_values, [0] * (len(ts) % 2)], axis=0
        )
        return netzero_shape_values


@comp_reg_deco
class Carrier(C3obj):
    """Represents the carrier of a pulse."""

    def __init__(
        self,
        name: str,
        desc: str = " ",
        comment: str = " ",
        params: dict = {},
    ):
        params_default = {
            "freq": Qty(value=0.0, min_val=-1.0, max_val=+1.0, unit="V"),
            "framechange": Qty(value=0.0, min_val=-np.pi, max_val=np.pi, unit="rad"),
        }
        params_default.update(params)
        super().__init__(
            name=name,
            desc=desc,
            comment=comment,
            params=params_default,
        )

    def write_config(self, filepath: str) -> None:
        """
        Write dictionary to a HJSON file.
        """
        with open(filepath, "w") as cfg_file:
            hjson.dump(self.asdict(), cfg_file, default=hjson_encode)

    def __repr__(self) -> str:
        repr_str = self.name + ":: "
        for key, item in self.params.items():
            repr_str += str(key) + " : " + str(item) + ", "
        return repr_str



"""Module for testing C3 Integration with Qiskit
"""

import json
from c3.libraries.constants import GATES
from c3.experiment import Experiment
from c3.c3objs import Quantity as Qty
from c3.qiskit import C3Provider
from c3.qiskit.c3_exceptions import C3QiskitError
from c3.qiskit.c3_gates import (
    RX90pGate,
    RX90mGate,
    RXpGate,
    RY90pGate,
    RY90mGate,
    RYpGate,
    RZ90pGate,
    RZ90mGate,
    RZpGate,
    CRXpGate,
    CRGate,
    CR90Gate,
    SetParamsGate,
)
from qiskit.circuit.library import RXGate, RYGate, RZGate, CRXGate
from qiskit.extensions import UnitaryGate
from qiskit.quantum_info import Statevector, Operator
from qiskit import transpile
from qiskit.providers import BackendV1 as Backend
from qiskit import execute, QuantumCircuit
import pytest
import numpy as np


@pytest.mark.unit
@pytest.mark.qiskit
def test_backends():
    """Test backends() function which returns all available backends"""
    c3_qiskit = C3Provider()
    for backend in c3_qiskit.backends():
        assert isinstance(backend, Backend)


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.parametrize(
    "backend", ["c3_qasm_perfect_simulator", "c3_qasm_physics_simulator"]
)
def test_get_backend(backend):
    """Test get_backend() which returns the backend with matching name

    Parameters
    ----------
    backend : str
        name of the backend that is to be fetched
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    assert isinstance(received_backend, Backend)


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.parametrize("backend", ["c3_qasm_perfect_simulator"])
def test_transpile(get_test_circuit, backend):  # noqa
    """Test the transpiling using our backends.
    Should throw error due to missing H gate

    Parameters
    ----------
    get_test_circuit : callable
        pytest fixture for a simple quantum circuit
    backend : str
        name of the backend which is to be tested
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    trans_qc = transpile(get_test_circuit, received_backend)
    assert Statevector.from_instruction(get_test_circuit).equiv(
        Statevector.from_instruction(trans_qc)
    )


@pytest.mark.integration
@pytest.mark.qiskit
@pytest.mark.slow
@pytest.mark.parametrize("backend", ["c3_qasm_perfect_simulator"])
def test_get_result(get_3_qubit_circuit, backend, get_result_qiskit):  # noqa
    """Test the counts from running a 3 qubit Circuit

    Parameters
    ----------
    get_3_qubit_circuit : callable
        pytest fixture for a 3 qubit circuit
    backend : str
        name of the backend which is to be tested
    simulation_type: str
        physics based or perfect gates simulation
    get_result_qiskit: callable
        pytest fixture for experiment results
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    received_backend.set_device_config("test/quickstart.hjson")
    received_backend.disable_flip_labels()
    qc = get_3_qubit_circuit
    job_sim = execute(qc, received_backend, shots=1000)
    result_sim = job_sim.result()

    # TODO: Test results with qiskit style qubit indexing
    # qiskit_simulator = Aer.get_backend("qasm_simulator")
    # qiskit_counts = execute(qc, qiskit_simulator, shots=1000).result().get_counts(qc)
    # assert result_sim.get_counts(qc) == qiskit_counts

    # Test results with original C3 style qubit indexing
    received_backend.disable_flip_labels()
    no_flip_counts = get_result_qiskit[backend]
    job_sim = execute(qc, received_backend, shots=1000)
    result_sim = job_sim.result()
    result_json = json.dumps(
        result_sim.to_dict()
    )  # ensure results can be properly serialised
    assert json.loads(result_json)["backend_name"] == backend
    assert result_sim.get_counts() == no_flip_counts


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.parametrize("backend", ["c3_qasm_perfect_simulator"])
def test_get_exception(get_bad_circuit, backend):  # noqa
    """Test to check exceptions

    Parameters
    ----------
    get_bad_circuit : callable
        pytest fixture for a circuit with conditional operation
    backend : str
        name of the backend which is to be tested
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    received_backend.set_device_config("test/quickstart.hjson")
    qc = get_bad_circuit

    with pytest.raises(C3QiskitError):
        execute(qc, received_backend, shots=1000)


@pytest.mark.integration
@pytest.mark.qiskit
@pytest.mark.slow
def test_qiskit_physics(get_physics_circuit):
    """API test for qiskit physics simulation"""
    c3_qiskit = C3Provider()
    physics_backend = c3_qiskit.get_backend("c3_qasm_physics_simulator")
    physics_backend.set_device_config("test/qiskit.cfg")
    qc = get_physics_circuit
    qc.measure_all()
    job_sim = physics_backend.run(qc)
    expected_pops = np.array([0, 0, 0.5, 0, 0, 0, 0.5, 0])
    received_pops = np.array(list(job_sim.result().data()["state_pops"].values()))
    np.testing.assert_allclose(received_pops, desired=expected_pops, atol=1e-1)


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.slow
def test_qiskit_parameter_update(get_physics_circuit):
    """Test for checking parameters are updated by the gate & options interface"""
    c3_qiskit = C3Provider()
    physics_backend = c3_qiskit.get_backend("c3_qasm_physics_simulator")
    physics_backend.set_device_config("test/qiskit.cfg")
    qc = get_physics_circuit  # TODO use a smaller circuit

    # Test that runtime options are correctly assigned
    opt_map = [[["rx90p[0]", "d1", "gaussian", "amp"]]]
    param = [
        Qty(value=0.5, min_val=0.2, max_val=0.6, unit="V"),
    ]
    _ = physics_backend.run(qc, params=param, opt_map=opt_map)
    physics_backend.c3_exp.pmap.set_opt_map(opt_map)
    np.testing.assert_allclose(
        physics_backend.c3_exp.pmap.get_parameter_dict()[
            "rx90p[0]-d1-gaussian-amp"
        ].numpy(),
        param[0].numpy(),
    )

    # Test that custom gate for setting parameters works
    amp = Qty(value=0.8, min_val=0.2, max_val=1, unit="V")
    opt_map = [[["rx90p[0]", "d1", "gaussian", "amp"]]]
    param_gate = SetParamsGate(params=[[amp.asdict()], opt_map])
    qc.append(param_gate, [0])
    _ = physics_backend.run(qc)
    np.testing.assert_allclose(
        physics_backend.c3_exp.pmap.get_parameter_dict()[
            "rx90p[0]-d1-gaussian-amp"
        ].numpy(),
        amp.numpy(),
    )

    # Test that SetParamsGate not at the end raises an error
    qc.append(RX90pGate(), [0])
    with pytest.raises(KeyError):
        _ = physics_backend.run(qc)


@pytest.mark.parametrize(
    ["backend", "config_file"],
    [
        pytest.param(
            "c3_qasm_perfect_simulator", "test/quickstart.hjson", id="perfect_sim"
        ),
        pytest.param("c3_qasm_physics_simulator", "test/qiskit.cfg", id="physics_sim"),
    ],
)
@pytest.mark.unit
@pytest.mark.qiskit
def test_too_many_qubits(backend, config_file):
    """Check that error is raised when circuit has more qubits than device

    Parameters
    ----------
    backend : tuple
        name and device config of the backend to be tested
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    received_backend.set_device_config(config_file)
    qc = QuantumCircuit(4, 4)
    with pytest.raises(C3QiskitError):
        execute(qc, received_backend, shots=1000)


@pytest.mark.parametrize(
    ["c3_gate", "c3_qubits", "qiskit_gate", "qiskit_qubits"],
    [
        pytest.param(RX90pGate(), [0], RXGate(theta=np.pi / 2.0), [0], id="rx90p"),
        pytest.param(RX90mGate(), [0], RXGate(theta=-np.pi / 2.0), [0], id="rx90m"),
        pytest.param(RXpGate(), [0], RXGate(theta=np.pi), [0], id="rxp"),
        pytest.param(RY90pGate(), [0], RYGate(theta=np.pi / 2.0), [0], id="ry90p"),
        pytest.param(RY90mGate(), [0], RYGate(theta=-np.pi / 2.0), [0], id="ry90m"),
        pytest.param(RYpGate(), [0], RYGate(theta=np.pi), [0], id="ryp"),
        pytest.param(RZ90pGate(), [0], RZGate(phi=np.pi / 2.0), [0], id="rz90p"),
        pytest.param(RZ90mGate(), [0], RZGate(phi=-np.pi / 2.0), [0], id="rz90m"),
        pytest.param(RZpGate(), [0], RZGate(phi=np.pi), [0], id="rzp"),
        # TODO Fix this dummy test for CRXp once matrix is resolved in c3_gates
        pytest.param(
            CRXpGate,
            [0, 1],
            CRXGate(theta=np.pi),
            [0, 1],
            marks=pytest.mark.xfail,
            id="crxp",
        ),
        pytest.param(CRGate(), [0, 1], UnitaryGate(data=GATES["cr"]), [0, 1], id="cr"),
        pytest.param(
            CR90Gate(), [0, 1], UnitaryGate(data=GATES["cr90"]), [0, 1], id="cr90"
        ),
    ],
)
@pytest.mark.unit
@pytest.mark.qiskit
def test_custom_c3_qiskit_gates(c3_gate, c3_qubits, qiskit_gate, qiskit_qubits):
    """Test custom c3 gates for qiskit have correct matrix representations

    Parameters
    ----------
    c3_gate : Gate
        A qiskit Gate object for c3 custom gates
    c3_qubits : List
        List containing the target qubits
    qiskit_gate : Gate
        Gate object for native qiskit gates
    qiskit_qubits : List
        List containing the target qubits
    """
    # TODO configure and test on c3 perfect simulator

    qc_c3 = QuantumCircuit(2, 2)
    qc_qiskit = QuantumCircuit(2, 2)

    qc_c3.append(c3_gate, c3_qubits)
    op_c3 = Operator(qc_c3)
    qc_qiskit.append(qiskit_gate, qiskit_qubits)
    op_qiskit = Operator(qc_qiskit)

    assert op_c3.equiv(op_qiskit)
    np.testing.assert_allclose(
        c3_gate.to_matrix(), desired=GATES[c3_gate.name], atol=1e-3
    )


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.parametrize(
    ["backend", "config_file"],
    [
        pytest.param(
            "c3_qasm_perfect_simulator", "test/quickstart.hjson", id="perfect_sim"
        ),
        pytest.param("c3_qasm_physics_simulator", "test/qiskit.cfg", id="physics_sim"),
    ],
)
def test_user_provided_c3_exp(backend, config_file):
    """Test for checking user provided C3 Experiment object is correctly assigned"""
    test_exp = Experiment()
    test_exp.load_quick_setup(config_file)
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    received_backend.set_c3_experiment(test_exp)
    assert received_backend.c3_exp is test_exp


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.parametrize(
    ["backend"],
    [
        pytest.param("c3_qasm_perfect_simulator", id="perfect_sim"),
        pytest.param("c3_qasm_physics_simulator", id="physics_sim"),
    ],
)
def test_experiment_not_initialised(backend, get_test_circuit):
    """Test for checking error is raised if c3 experiment object is not correctly initialised"""
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    qc = get_test_circuit
    with pytest.raises(C3QiskitError):
        received_backend.run(qc)


@pytest.mark.qiskit
@pytest.mark.unit
def test_initial_statevector(get_physics_circuit):
    """Check initial statevector is correctly validated

    Parameters
    ----------
    get_physics_circuit : QuantumCircuit
        Circuit fixture with RX90p and CR90 gates
    """
    c3_qiskit = C3Provider()
    physics_backend = c3_qiskit.get_backend("c3_qasm_physics_simulator")
    physics_backend.set_device_config("test/qiskit.cfg")
    qc = get_physics_circuit
    qc.measure_all()
    with pytest.raises(C3QiskitError):
        # incorrect dimension
        physics_backend.run(qc, initial_statevector=[0, 0, 0, 0, 0, 0, 1])

    with pytest.raises(C3QiskitError):
        # unnormalised
        physics_backend.run(qc, initial_statevector=[1, 1, 1, 1, 1, 1, 1, 1])



"""
integration testing module for C1 optimization through two-qubits example
"""

import pickle
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal as almost_equal

# Libs and helpers
from c3.libraries.propagation import rk4

with open("test/two_qubit_data.pickle", "rb") as filename:
    test_data = pickle.load(filename)


@pytest.mark.tensorflow
@pytest.mark.unit
def test_signals(get_two_qubit_chip) -> None:
    """Compare generated control signals to stored examples."""
    exp = get_two_qubit_chip
    pmap = exp.pmap
    gen_signal = pmap.generator.generate_signals(pmap.instructions["rx90p[0]"])
    ts = gen_signal["d1"]["ts"]
    np.testing.assert_allclose(ts, test_data["ts"])
    np.testing.assert_allclose(
        actual=gen_signal["d1"]["values"].numpy(),
        desired=test_data["signal"]["d1"]["values"].numpy(),
    )


@pytest.mark.unit
def test_hamiltonians(get_two_qubit_chip) -> None:
    """Compare Hamilonians"""
    model = get_two_qubit_chip.pmap.model
    hdrift, hks = model.get_Hamiltonians()
    assert (hdrift.numpy() - test_data["hdrift"].numpy() < 1).any()
    for key in hks:
        almost_equal(hks[key], test_data["hks"][key])


@pytest.mark.tensorflow
@pytest.mark.integration
def test_propagation(get_two_qubit_chip) -> None:
    """Test that result of the propagation code does not change."""
    GATE_STR = "rx90p[0]"
    exp = get_two_qubit_chip
    pmap = exp.pmap
    instr = pmap.instructions[GATE_STR]
    steps = int((instr.t_end - instr.t_start) * exp.sim_res)
    result = exp.propagation(
        pmap.model,
        pmap.generator,
        pmap.instructions["rx90p[0]"],
        exp.folding_stack[steps],
    )
    propagator = result["U"]
    almost_equal(propagator, test_data["propagator"])


@pytest.mark.unit
def test_init_point(get_OC_optimizer) -> None:
    """Check that a previous best point can be loaded as an initial point."""
    opt = get_OC_optimizer
    opt.load_best("test/best_point_open_loop.c3log")


@pytest.mark.slow
@pytest.mark.tensorflow
@pytest.mark.optimizers
@pytest.mark.integration
def test_optim(get_OC_optimizer) -> None:
    """
    check if optimization result is below 1e-1
    """
    opt = get_OC_optimizer
    opt.optimize_controls()
    assert opt.current_best_goal < 0.1


@pytest.mark.tensorflow
@pytest.mark.integration
def test_rk4(get_two_qubit_chip) -> None:
    """Testing that RK4 exists and runs."""
    exp = get_two_qubit_chip
    pmap = exp.pmap
    exp.set_prop_method(rk4)
    exp.propagation(pmap.model, pmap.generator, pmap.instructions["rx90p[0]"])



"""Collection of estimator functions, to compare two sets of (noisy) data."""

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp


estimators = dict()


def estimator_reg_deco(func):
    """
    Decorator for making registry of functions
    """
    estimators[str(func.__name__)] = func
    return func


@estimator_reg_deco
def mean_dist(exp_values, sim_values, exp_stds, shots):
    """Return the root mean squared of the differences."""
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tf.reduce_mean(diffs)


@estimator_reg_deco
def median_dist(exp_values, sim_values, exp_stds, shots):
    """Return the median of the differences."""
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tfp.stats.percentile(diffs, 50.0, interpolation="midpoint")


@estimator_reg_deco
def rms_dist(exp_values, sim_values, exp_stds, shots):
    """Return the root mean squared of the differences."""
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tf.sqrt(tf.reduce_mean(diffs**2))


@estimator_reg_deco
def mean_sim_stds_dist(exp_values, sim_values, exp_stds, shots):
    """Return the mean of the distance in number of exp_stds away."""
    sim_std = tf.sqrt(sim_values * (1 - sim_values) / shots)
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tf.reduce_mean(diffs / sim_std)


@estimator_reg_deco
def rms_sim_stds_dist(exp_values, sim_values, exp_stds, shots):
    """Return the root mean squared of the differences measured in exp_stds."""
    sim_std = tf.sqrt(sim_values * (1 - sim_values) / shots)
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tf.sqrt(tf.reduce_mean((diffs / sim_std) ** 2))


@estimator_reg_deco
def mean_exp_stds_dist(exp_values, sim_values, exp_stds, shots):
    """Return the mean of the distance in number of exp_stds away."""
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tf.reduce_mean(diffs / exp_stds)


@estimator_reg_deco
def rms_exp_stds_dist(exp_values, sim_values, exp_stds, shots):
    """Return the root mean squared of the differences measured in exp_stds."""
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tf.sqrt(tf.reduce_mean((diffs / exp_stds) ** 2))


@estimator_reg_deco
def std_of_diffs(exp_values, sim_values, exp_stds, shots):
    """Return the std of the distances."""
    diffs = tf.abs(tf.subtract(exp_values, sim_values))
    return tf.math.reduce_std(diffs)


@estimator_reg_deco
def neg_loglkh_binom(exp_values, sim_values, exp_stds, shots):
    """
    Average likelihood of the experimental values with binomial distribution.

    Return the likelihood of the experimental values given the simulated
    values, and given a binomial distribution function.
    """
    binom = tfp.distributions.Binomial(total_count=shots, probs=sim_values)
    loglkhs = binom.log_prob(exp_values * shots)
    loglkh = tf.reduce_mean(loglkhs)
    return -loglkh


@estimator_reg_deco
def neg_loglkh_binom_norm(exp_values, sim_values, exp_stds, shots):
    """
    Average likelihood of the exp values with normalised binomial distribution.

    Return the likelihood of the experimental values given the simulated
    values, and given a binomial distribution function that is normalised to
    give probability 1 at the top of the distribution.
    """

    binom = tfp.distributions.Binomial(total_count=shots, probs=sim_values)
    loglkhs = binom.log_prob(exp_values * shots) - binom.log_prob(sim_values * shots)
    loglkh = tf.reduce_mean(loglkhs)
    return -loglkh


@estimator_reg_deco
def neg_loglkh_gauss(exp_values, sim_values, exp_stds, shots):
    """
    Likelihood of the experimental values.

    The distribution is assumed to be binomial (approximated by a gaussian).
    """
    std = tf.sqrt(sim_values * (1 - sim_values) / shots)
    mean = sim_values
    gauss = tfp.distributions.Normal(mean, std)
    loglkhs = gauss.log_prob(exp_values)
    loglkh = tf.reduce_mean(loglkhs)
    return -loglkh


@estimator_reg_deco
def neg_loglkh_gauss_norm(exp_values, sim_values, exp_stds, shots):
    """
    Likelihood of the experimental values.

    The distribution is assumed to be binomial (approximated by a gaussian)
    that is normalised to give probability 1 at the top of the distribution.
    """
    std = tf.sqrt(sim_values * (1 - sim_values) / shots)
    mean = sim_values
    gauss = tfp.distributions.Normal(mean, std)
    loglkhs = gauss.log_prob(exp_values) - gauss.log_prob(mean)
    loglkh = tf.reduce_mean(loglkhs)
    return -loglkh


@estimator_reg_deco
def neg_loglkh_gauss_norm_sum(exp_values, sim_values, exp_stds, shots):
    """
    Likelihood of the experimental values.

    The distribution is assumed to be binomial (approximated by a gaussian)
    that is normalised to give probability 1 at the top of the distribution.
    """
    std = tf.sqrt(sim_values * (1 - sim_values) / shots)
    mean = sim_values
    gauss = tfp.distributions.Normal(mean, std)
    loglkhs = gauss.log_prob(exp_values) - gauss.log_prob(mean)
    loglkh = tf.reduce_sum(loglkhs)
    return -loglkh


@estimator_reg_deco
def g_LL_prime(exp_values, sim_values, exp_stds, shots):
    std = tf.sqrt(sim_values * (1 - sim_values) / shots)
    return tf.reduce_mean(((exp_values - sim_values) ** 2 / std**2 - 1) / 2)


def dv_g_LL_prime(gs, dv_gs, weights):
    K = np.sum(weights)
    g = 0
    for ii in range(len(weights)):
        g += weights[ii] * dv_gs[ii]
    return g / K


def g_LL_prime_combined(gs, weights):
    K = np.sum(weights)
    return np.sum(np.array(weights) * gs) / K


@estimator_reg_deco
def neg_loglkh_multinom(exp_values, sim_values, exp_stds, shots):
    """
    Average likelihood of the experimental values with multinomial distribution.

    Return the likelihood of the experimental values given the simulated
    values, and given a multinomial distribution function.
    """
    multi = tfp.distributions.Multinomial(
        total_count=tf.reshape(shots, [shots.shape[0]]), probs=sim_values
    )
    loglkhs = multi.log_prob(exp_values * shots)
    loglkh = tf.reduce_mean(loglkhs)
    return -loglkh


@estimator_reg_deco
def neg_loglkh_multinom_norm(exp_values, sim_values, exp_stds, shots):
    """
    Average likelihood of the experimental values with multinomial distribution.

    Return the likelihood of the experimental values given the simulated
    values, and given a multinomial distribution function that is normalised to
    give probability 1 at the top of the distribution.
    """
    multi = tfp.distributions.Multinomial(
        total_count=tf.reshape(shots, [shots.shape[0]]), probs=sim_values
    )
    loglkhs = multi.log_prob(exp_values * shots) - multi.log_prob(sim_values * shots)
    loglkh = tf.reduce_mean(loglkhs)
    return -loglkh - 1.5



"""
Experiment class that models and simulates the whole experiment.

It combines the information about the model of the quantum device, the control stack
and the operations that can be done on the device.

Given this information an experiment run is simulated, returning either processes,
states or populations.
"""

import os
import copy
import pickle
import itertools
import hjson
import numpy as np
import tensorflow as tf
from typing import Dict, List
import time

from c3.c3objs import hjson_encode, hjson_decode
from c3.generator.generator import Generator
from c3.parametermap import ParameterMap
from c3.signal.gates import Instruction
from c3.model import Model
from c3.utils.tf_utils import (
    tf_state_to_dm,
    tf_super,
    tf_vec_to_dm,
    _tf_matmul_n_even,
    _tf_matmul_n_odd,
)

from c3.libraries.propagation import unitary_provider, state_provider

from c3.utils.qt_utils import perfect_single_q_parametric_gate


class Experiment:
    """
    It models all of the behaviour of the physical experiment, serving as a
    host for the individual parts making up the experiment.

    Parameters
    ----------
    pmap: ParameterMap
        including
        model: Model
            The underlying physical device.
        generator: Generator
            The infrastructure for generating and sending control signals to the
            device.
        gateset: GateSet
            A gate level description of the operations implemented by control
            pulses.

    """

    def __init__(self, pmap: ParameterMap = None, prop_method=None, sim_res=100e9):
        self.pmap = pmap
        self.opt_gates = None
        self.propagators: Dict[str, tf.Tensor] = {}
        self.partial_propagators: Dict = {}
        self.created_by = None
        self.logdir: str = ""
        self.propagate_batch_size = None
        self.use_control_fields = True
        self.overwrite_propagators = True  # Keep only currently computed propagators
        self.compute_propagators_timestamp = 0
        self.stop_partial_propagator_gradient = True
        self.evaluate = self.evaluate_legacy
        self.sim_res = sim_res
        self.set_prop_method(prop_method)

    def set_prop_method(self, prop_method=None) -> None:
        """
        Configure the selected propagation method by either linking the function handle or
        looking it up in the library.
        """
        if prop_method is None:
            self.propagation = unitary_provider["pwc"]
            if self.pmap is not None:
                self._compute_folding_stack()
        elif isinstance(prop_method, str):
            try:
                self.propagation = unitary_provider[prop_method]
            except KeyError:
                self.propagation = state_provider[prop_method]
        elif callable(prop_method):
            self.propagation = prop_method

    def _compute_folding_stack(self):
        self.folding_stack = {}
        for instr in self.pmap.instructions.values():
            n_steps = int((instr.t_end - instr.t_start) * self.sim_res)
            if n_steps not in self.folding_stack:
                stack = []
                while n_steps > 1:
                    if not n_steps % 2:  # is divisable by 2
                        stack.append(_tf_matmul_n_even)
                    else:
                        stack.append(_tf_matmul_n_odd)
                    n_steps = np.ceil(n_steps / 2)
                self.folding_stack[
                    int((instr.t_end - instr.t_start) * self.sim_res)
                ] = stack

    def enable_qasm(self) -> None:
        """
        Switch the sequencing format to QASM. Will become the default.
        """
        self.evaluate = self.evaluate_qasm

    def set_created_by(self, config):
        """
        Store the config file location used to created this experiment.
        """

        self.created_by = config

    def load_quick_setup(self, filepath: str) -> None:
        """
        Load a quick setup file.

        Parameters
        ----------
        filepath : str
            Location of the configuration file

        """
        with open(filepath, "r") as cfg_file:
            cfg = hjson.loads(cfg_file.read(), object_pairs_hook=hjson_decode)
        self.quick_setup(cfg, os.path.dirname(filepath))

    def quick_setup(self, cfg, base_dir: str = None) -> None:
        """
        Load a quick setup cfg and create all necessary components.

        Parameters
        ----------
        cfg : Dict
            Configuration options

        """

        def make_absolute(filename: str) -> str:
            if base_dir or os.path.isabs(filename):
                return os.path.join(base_dir, filename)
            else:
                return filename

        model = Model()
        model.read_config(make_absolute(cfg["model"]))
        gen = Generator()
        gen.read_config(make_absolute(cfg["generator"]))

        single_gate_time = cfg["single_qubit_gate_time"]
        v2hz = cfg["v2hz"]
        instructions = []
        sideband = cfg.pop("sideband", None)
        for gate_name, props in cfg["single_qubit_gates"].items():
            target_qubit = model.subsystems[props["qubits"]]
            instr = Instruction(
                name=props["name"],
                targets=[model.names.index(props["qubits"])],
                t_start=0.0,
                t_end=single_gate_time,
                channels=[target_qubit.drive_line],
            )
            instr.quick_setup(
                target_qubit.drive_line,
                target_qubit.params["freq"].get_value() / 2 / np.pi,
                single_gate_time,
                v2hz,
                sideband,
            )
            instructions.append(instr)

        for gate_name, props in cfg["two_qubit_gates"].items():
            qubit_1 = model.subsystems[props["qubit_1"]]
            qubit_2 = model.subsystems[props["qubit_2"]]
            instr = Instruction(
                name=props["name"],
                targets=[
                    model.names.index(props["qubit_1"]),
                    model.names.index(props["qubit_2"]),
                ],
                t_start=0.0,
                t_end=props["gate_time"],
                channels=[qubit_1.drive_line, qubit_2.drive_line],
            )
            instr.quick_setup(
                qubit_1.drive_line,
                qubit_1.params["freq"].get_value() / 2 / np.pi,
                props["gate_time"],
                v2hz,
                sideband,
            )
            instr.quick_setup(
                qubit_2.drive_line,
                qubit_2.params["freq"].get_value() / 2 / np.pi,
                props["gate_time"],
                v2hz,
                sideband,
            )
            instructions.append(instr)

        self.sim_res = 100e9
        self.pmap = ParameterMap(instructions, generator=gen, model=model)
        self.set_prop_method()

    def read_config(self, filepath: str) -> None:
        """
        Load a file and parse it to create a Model object.

        Parameters
        ----------
        filepath : str
            Location of the configuration file

        """
        with open(filepath, "r") as cfg_file:
            cfg = hjson.loads(cfg_file.read(), object_pairs_hook=hjson_decode)
        self.from_dict(cfg)

    def from_dict(self, cfg: Dict) -> None:
        """
        Load experiment from dictionary
        """
        model = Model()
        model.fromdict(cfg["model"])
        generator = Generator()
        generator.fromdict(cfg["generator"])
        pmap = ParameterMap(model=model, generator=generator)
        pmap.fromdict(cfg["instructions"])
        if "options" in cfg:
            for k, v in cfg["options"].items():
                self.__dict__[k] = v
        self.pmap = pmap
        self.sim_res = cfg.pop("sim_res", 100e9)
        self.set_prop_method()

    def write_config(self, filepath: str) -> None:
        """
        Write dictionary to a HJSON file.
        """
        with open(filepath, "w") as cfg_file:
            hjson.dump(self.asdict(), cfg_file, default=hjson_encode)

    def asdict(self) -> Dict:
        """
        Return a dictionary compatible with config files.
        """
        exp_dict: Dict[str, Dict] = {}
        exp_dict["instructions"] = {}
        for name, instr in self.pmap.instructions.items():
            exp_dict["instructions"][name] = instr.asdict()
        exp_dict["model"] = self.pmap.model.asdict()
        exp_dict["generator"] = self.pmap.generator.asdict()
        exp_dict["options"] = {
            "propagate_batch_size": self.propagate_batch_size,
            "use_control_fields": self.use_control_fields,
            "overwrite_propagators": self.overwrite_propagators,
            "stop_partial_propagator_gradient": self.stop_partial_propagator_gradient,
        }
        exp_dict["sim_res"] = self.sim_res
        return exp_dict

    def __str__(self) -> str:
        return hjson.dumps(self.asdict(), default=hjson_encode)

    def evaluate_legacy(self, sequences, psi_init: tf.Tensor = None):
        """
        Compute the population values for a given sequence of operations.

        Parameters
        ----------
        sequences: str list
            A list of control pulses/gates to perform on the device.

        psi_init: tf.Tensor
            A tensor containing the initial statevector

        Returns
        -------
        list
            A list of populations

        """
        model = self.pmap.model
        if psi_init is None:
            psi_init = model.get_init_state()
        populations = []
        for sequence in sequences:
            psi_t = copy.deepcopy(psi_init)
            for gate in sequence:
                psi_t = tf.matmul(self.propagators[gate], psi_t)

            pops = self.populations(psi_t, model.lindbladian)
            populations.append(pops)
        return populations

    def evaluate_qasm(self, sequences, psi_init: tf.Tensor = None):
        """
        Compute the population values for a given sequence (in QASM format) of
        operations.

        Parameters
        ----------
        sequences: dict list
            A list of control pulses/gates to perform on the device in QASM format.

        psi_init: tf.Tensor
            A tensor containing the initial statevector

        Returns
        -------
        list
            A list of populations

        """
        model = self.pmap.model
        if psi_init is None:
            psi_init = model.get_init_state()
        self.psi_init = psi_init
        populations = []
        for sequence in sequences:
            psi_t = copy.deepcopy(self.psi_init)
            for gate in sequence:
                psi_t = tf.matmul(self.lookup_gate(**gate), psi_t)

            pops = self.populations(psi_t, model.lindbladian)
            populations.append(pops)
        return populations

    def lookup_gate(self, name, qubits, params=None) -> tf.constant:
        """
        Returns a fixed operation or a parametric virtual Z gate. To be extended to
        general parametric gates.
        """
        if name == "VZ":
            gate = tf.constant(self.get_VZ(qubits, params))
        else:
            gate = self.propagators[name + str(qubits)]
        return gate

    def get_VZ(self, target, params):
        """
        Returns the appropriate Z-rotation.
        """
        dims = self.pmap.model.dims
        return perfect_single_q_parametric_gate("Z", target[0], params[0], dims)

    def process(self, populations, labels=None):
        """
        Apply a readout procedure to a population vector. Very specialized
        at the moment.

        Parameters
        ----------
        populations: list
            List of populations from evaluating.

        labels: list
            List of state labels specifying a subspace.

        Returns
        -------
        list
            A list of processed populations.

        """
        model = self.pmap.model
        populations_final = []
        populations_no_rescale = []
        for pops in populations:
            # TODO: Loop over all model.tasks in a general fashion
            # TODO: Selecting states by label in the case of computational space
            if "conf_matrix" in model.tasks:
                pops = model.tasks["conf_matrix"].confuse(pops)
                if labels is not None:
                    pops_select = 0
                    for label in labels:
                        pops_select += pops[model.comp_state_labels.index(label)]
                    pops = pops_select
                else:
                    pops = tf.reshape(pops, [pops.shape[0]])
            else:
                if labels is not None:
                    pops_select = 0
                    for label in labels:
                        try:
                            pops_select += pops[model.state_labels.index(label)]
                        except ValueError:
                            raise Exception(
                                f"C3:ERROR:State {label} not defined. Available are:\n"
                                f"{model.state_labels}"
                            )
                    pops = pops_select
                else:
                    pops = tf.reshape(pops, [pops.shape[0]])
            if "meas_rescale" in model.tasks:
                populations_no_rescale.append(pops)
                pops = model.tasks["meas_rescale"].rescale(pops)
            populations_final.append(pops)
        return populations_final, populations_no_rescale

    def get_perfect_gates(self, gate_keys: list = None) -> Dict[str, np.ndarray]:
        """Return a perfect gateset for the gate_keys.

        Parameters
        ----------
        gate_keys: list
            (Optional) List of gates to evaluate.

        Returns
        -------
        Dict[str, np.array]
            A dictionary of gate names and np.array representation
            of the corresponding unitary

        Raises
        ------
        Exception
            Raise general exception for undefined gate
        """
        instructions = self.pmap.instructions
        gates = {}
        dims = self.pmap.model.dims
        if gate_keys is None:
            gate_keys = instructions.keys()  # type: ignore
        for gate in gate_keys:
            gates[gate] = instructions[gate].get_ideal_gate(dims)

        # TODO parametric gates

        return gates

    def compute_states(self) -> Dict[Instruction, List[tf.Tensor]]:
        """Employ a state solver to compute the trajectory of the system.

        Returns
        -------
        List[tf.tensor]
            List of states of the system from simulation.

        """
        model = self.pmap.model
        generator = self.pmap.generator
        instructions = self.pmap.instructions
        states = {}

        gate_ids = self.opt_gates
        if gate_ids is None:
            gate_ids = instructions.keys()

        for gate in gate_ids:
            try:
                instr = instructions[gate]
            except KeyError:
                raise Exception(
                    f"C3:Error: Gate '{gate}' is not defined."
                    f" Available gates are:\n {list(instructions.keys())}."
                )
            signal = generator.generate_signals(instr)
            result = self.propagation(model, signal, self.folding_stack)
            states[instr] = result["states"]
        self.states = states
        return result

    def compute_propagators(self):
        """
        Compute the unitary representation of operations. If no operations are
        specified in self.opt_gates the complete gateset is computed.

        Returns
        -------
        dict
            A dictionary of gate names and their unitary representation.
        """
        model = self.pmap.model
        generator = self.pmap.generator
        instructions = self.pmap.instructions
        propagators = {}
        partial_propagators = {}
        gate_ids = self.opt_gates
        if gate_ids is None:
            gate_ids = instructions.keys()

        for gate in gate_ids:
            try:
                instr = instructions[gate]
            except KeyError:
                raise Exception(
                    f"C3:Error: Gate '{gate}' is not defined."
                    f" Available gates are:\n {list(instructions.keys())}."
                )

            model.controllability = self.use_control_fields
            steps = int((instr.t_end - instr.t_start) * self.sim_res)
            result = self.propagation(
                model, generator, instr, self.folding_stack[steps]
            )
            U = result["U"]
            dUs = result["dUs"]
            self.ts = result["ts"]
            if model.use_FR:
                # TODO change LO freq to at the level of a line
                freqs = {}
                framechanges = {}
                for line, ctrls in instr.comps.items():
                    # TODO calculate properly the average frequency that each qubit sees
                    offset = tf.constant(0.0, tf.float64)
                    for ctrl in ctrls.values():
                        if "freq_offset" in ctrl.params.keys():
                            if ctrl.params["amp"] != 0.0:
                                offset = ctrl.params["freq_offset"].get_value()
                    freqs[line] = tf.cast(
                        ctrls["carrier"].params["freq"].get_value() + offset,
                        tf.complex128,
                    )
                    framechanges[line] = tf.cast(
                        ctrls["carrier"].params["framechange"].get_value(),
                        tf.complex128,
                    )
                t_final = tf.constant(instr.t_end - instr.t_start, dtype=tf.complex128)
                FR = model.get_Frame_Rotation(t_final, freqs, framechanges)
                if model.lindbladian:
                    SFR = tf_super(FR)
                    U = tf.matmul(SFR, U)
                    self.FR = SFR
                else:
                    U = tf.matmul(FR, U)
                    self.FR = FR
            if model.dephasing_strength != 0.0:
                if not model.lindbladian:
                    raise ValueError("Dephasing can only be added when lindblad is on.")
                else:
                    amps = {}
                    for line, ctrls in instr.comps.items():
                        amp, sum = generator.devices["awg"].get_average_amp()
                        amps[line] = tf.cast(amp, tf.complex128)
                    t_final = tf.constant(
                        instr.t_end - instr.t_start, dtype=tf.complex128
                    )
                    dephasing_channel = model.get_dephasing_channel(t_final, amps)
                    U = tf.matmul(dephasing_channel, U)
            propagators[gate] = U
            partial_propagators[gate] = dUs

        # TODO we might want to move storing of the propagators to the instruction object
        if self.overwrite_propagators:
            self.propagators = propagators
            self.partial_propagators = partial_propagators
        else:
            self.propagators.update(propagators)
            self.partial_propagators.update(partial_propagators)
        self.compute_propagators_timestamp = time.time()
        return propagators

    def set_opt_gates(self, gates):
        """
        Specify a selection of gates to be computed.

        Parameters
        ----------
        gates: Identifiers of the gates of interest. Can contain duplicates.

        """
        if type(gates) is str:
            gates = [gates]
        self.opt_gates = gates

    def set_opt_gates_seq(self, seqs):
        """
        Specify a selection of gates to be computed.

        Parameters
        ----------
        seqs: Identifiers of the sequences of interest. Can contain duplicates.

        """
        self.opt_gates = list(set(itertools.chain.from_iterable(seqs)))

    def set_enable_store_unitaries(self, flag, logdir, exist_ok=False):
        """
        Saving of unitary propagators.

        Parameters
        ----------
        flag: boolean
            Enable or disable saving.
        logdir: str
            File path location for the resulting unitaries.
        """
        self.enable_store_unitaries = flag
        self.logdir = logdir
        if self.enable_store_unitaries:
            os.makedirs(self.logdir + "unitaries/", exist_ok=exist_ok)
            self.store_unitaries_counter = 0

    def store_Udict(self, goal):
        """
        Save unitary as text and pickle.

        Parameter
        ---------
        goal: tf.float64
            Value of the goal function, if used.

        """
        folder = (
            self.logdir
            + "unitaries/eval_"
            + str(self.store_unitaries_counter)
            + "_"
            + str(goal)
            + "/"
        )
        if not os.path.exists(folder):
            os.mkdir(folder)
        with open(folder + "Us.pickle", "wb+") as file:
            pickle.dump(self.propagators, file)
        for key, value in self.propagators.items():
            # Windows is not able to parse ":" as file path
            np.savetxt(folder + key.replace(":", ".") + ".txt", value)

    def populations(self, state, lindbladian):
        """
        Compute populations from a state or density vector.

        Parameters
        ----------
        state: tf.Tensor
            State or densitiy vector.
        lindbladian: boolean
            Specify if conversion to density matrix is needed.

        Returns
        -------
        tf.Tensor
            Vector of populations.
        """
        if lindbladian:
            rho = tf_vec_to_dm(state)
            pops = tf.math.real(tf.linalg.diag_part(rho))
            return tf.reshape(pops, shape=[pops.shape[0], 1])
        else:
            return tf.abs(state) ** 2

    def expect_oper(self, state, lindbladian, oper):
        if lindbladian:
            rho = tf_vec_to_dm(state)
        else:
            rho = tf_state_to_dm(state)
        trace = np.trace(np.matmul(rho, oper))
        return [[np.real(trace)]]  # ,[np.imag(trace)]]
